from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class ProcessUpdateRewardRateArgs(typing.TypedDict):
    scorevars_bump: int
    scorevars_ship_bump: int
    new_reward_rate_per_second: int


layout = borsh.CStruct(
    "scorevars_bump" / borsh.U8,
    "scorevars_ship_bump" / borsh.U8,
    "new_reward_rate_per_second" / borsh.U64,
)


class ProcessUpdateRewardRateAccounts(typing.TypedDict):
    update_authority_account: PublicKey
    score_vars_account: PublicKey
    score_vars_ship_account: PublicKey
    ship_mint: PublicKey


def process_update_reward_rate(
    args: ProcessUpdateRewardRateArgs, accounts: ProcessUpdateRewardRateAccounts
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["update_authority_account"],
            is_signer=True,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["score_vars_account"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["score_vars_ship_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["ship_mint"], is_signer=False, is_writable=False),
    ]
    identifier = b"\x8f\xab>\xd0\xe3\xd7E\x12"
    encoded_args = layout.build(
        {
            "scorevars_bump": args["scorevars_bump"],
            "scorevars_ship_bump": args["scorevars_ship_bump"],
            "new_reward_rate_per_second": args["new_reward_rate_per_second"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



import typing
import re
from solana.rpc.core import RPCException
from ..program_id import PROGRAM_ID
from . import anchor
from . import custom


def from_code(code: int) -> typing.Union[custom.CustomError, anchor.AnchorError, None]:
    return custom.from_code(code) if code >= 6000 else anchor.from_code(code)


error_re = re.compile(r"Program (\w+) failed: custom program error: (\w+)")


def _find_first_match(logs: list[str]) -> typing.Optional[re.Match]:
    for logline in logs:
        first_match = error_re.match(logline)
        if first_match is not None:
            return first_match
    return None


def from_tx_error(
    error: RPCException,
) -> typing.Union[anchor.AnchorError, custom.CustomError, None]:
    err_info = error.args[0]
    if "data" not in err_info:
        return None
    if "logs" not in err_info["data"]:
        return None
    first_match = _find_first_match(err_info["data"]["logs"])
    if first_match is None:
        return None
    program_id_raw, code_raw = first_match.groups()
    if program_id_raw != str(PROGRAM_ID):
        return None
    try:
        error_code = int(code_raw, 16)
    except ValueError:
        return None
    return from_code(error_code)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class UpdateCurrencyVaultAccounts(typing.TypedDict):
    update_authority_account: PublicKey
    market_vars_account: PublicKey
    registered_currency: PublicKey
    currency_mint: PublicKey
    sa_currency_vault: PublicKey
    system_program: PublicKey


def update_currency_vault(
    accounts: UpdateCurrencyVaultAccounts,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["update_authority_account"],
            is_signer=True,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["market_vars_account"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["registered_currency"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["currency_mint"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["sa_currency_vault"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
    ]
    identifier = b"\x12\x88H\x1fL\xf2\nR"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



from .deregister_currency import deregister_currency, DeregisterCurrencyAccounts
from .initialize_marketplace import (
    initialize_marketplace,
    InitializeMarketplaceAccounts,
)
from .initialize_open_orders_counter import (
    initialize_open_orders_counter,
    InitializeOpenOrdersCounterAccounts,
)
from .process_initialize_buy import (
    process_initialize_buy,
    ProcessInitializeBuyArgs,
    ProcessInitializeBuyAccounts,
)
from .process_initialize_sell import (
    process_initialize_sell,
    ProcessInitializeSellArgs,
    ProcessInitializeSellAccounts,
)
from .process_exchange import (
    process_exchange,
    ProcessExchangeArgs,
    ProcessExchangeAccounts,
)
from .process_cancel import process_cancel, ProcessCancelAccounts
from .register_currency import (
    register_currency,
    RegisterCurrencyArgs,
    RegisterCurrencyAccounts,
)
from .update_currency_vault import update_currency_vault, UpdateCurrencyVaultAccounts
from .update_currency_royalty import (
    update_currency_royalty,
    UpdateCurrencyRoyaltyArgs,
    UpdateCurrencyRoyaltyAccounts,
)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class RoyaltyTierJSON(typing.TypedDict):
    stake_amount: int
    discount: int


@dataclass
class RoyaltyTier:
    layout: typing.ClassVar = borsh.CStruct(
        "stake_amount" / borsh.U64, "discount" / borsh.U64
    )
    stake_amount: int
    discount: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "RoyaltyTier":
        return cls(stake_amount=obj.stake_amount, discount=obj.discount)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"stake_amount": self.stake_amount, "discount": self.discount}

    def to_json(self) -> RoyaltyTierJSON:
        return {"stake_amount": self.stake_amount, "discount": self.discount}

    @classmethod
    def from_json(cls, obj: RoyaltyTierJSON) -> "RoyaltyTier":
        return cls(stake_amount=obj["stake_amount"], discount=obj["discount"])



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class ProcessCancelAccounts(typing.TypedDict):
    order_initializer: PublicKey
    deposit_mint: PublicKey
    initializer_deposit_token_account: PublicKey
    order_vault_account: PublicKey
    order_vault_authority: PublicKey
    order_account: PublicKey
    open_orders_counter: PublicKey
    token_program: PublicKey


def process_cancel(accounts: ProcessCancelAccounts) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["order_initializer"], is_signer=True, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["deposit_mint"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["initializer_deposit_token_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["order_vault_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["order_vault_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["order_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["open_orders_counter"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["token_program"], is_signer=False, is_writable=False
        ),
    ]
    identifier = b"UT\xd6\xf0\x8c)\xe6\x95"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class AssetJSON(typing.TypedDict):
    kind: typing.Literal["Asset"]


class CurrencyJSON(typing.TypedDict):
    kind: typing.Literal["Currency"]


@dataclass
class Asset:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Asset"

    @classmethod
    def to_json(cls) -> AssetJSON:
        return AssetJSON(
            kind="Asset",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Asset": {},
        }


@dataclass
class Currency:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Currency"

    @classmethod
    def to_json(cls) -> CurrencyJSON:
        return CurrencyJSON(
            kind="Currency",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Currency": {},
        }


TokenTypeKind = typing.Union[Asset, Currency]
TokenTypeJSON = typing.Union[AssetJSON, CurrencyJSON]


def from_decoded(obj: dict) -> TokenTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Asset" in obj:
        return Asset()
    if "Currency" in obj:
        return Currency()
    raise ValueError("Invalid enum object")


def from_json(obj: TokenTypeJSON) -> TokenTypeKind:
    if obj["kind"] == "Asset":
        return Asset()
    if obj["kind"] == "Currency":
        return Currency()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Asset" / borsh.CStruct(), "Currency" / borsh.CStruct())



import typing
from . import royalty_tier
from .royalty_tier import RoyaltyTier, RoyaltyTierJSON
from . import order_side
from .order_side import OrderSideKind, OrderSideJSON
from . import token_type
from .token_type import TokenTypeKind, TokenTypeJSON



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class UpdateCurrencyRoyaltyArgs(typing.TypedDict):
    royalty: int


layout = borsh.CStruct("royalty" / borsh.U64)


class UpdateCurrencyRoyaltyAccounts(typing.TypedDict):
    update_authority_account: PublicKey
    market_vars_account: PublicKey
    registered_currency: PublicKey
    currency_mint: PublicKey
    system_program: PublicKey


def update_currency_royalty(
    args: UpdateCurrencyRoyaltyArgs, accounts: UpdateCurrencyRoyaltyAccounts
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["update_authority_account"],
            is_signer=True,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["market_vars_account"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["registered_currency"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["currency_mint"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
    ]
    identifier = b"\xb3\xe8\x05*\xccZ\xae\xf8"
    encoded_args = layout.build(
        {
            "royalty": args["royalty"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class ProcessCloseAccountsArgs(typing.TypedDict):
    staking_bump: int
    scorevars_bump: int
    ship_bump: int
    fuel_bump: int
    food_bump: int
    arms_bump: int
    escrow_auth_bump: int


layout = borsh.CStruct(
    "staking_bump" / borsh.U8,
    "scorevars_bump" / borsh.U8,
    "ship_bump" / borsh.U8,
    "fuel_bump" / borsh.U8,
    "food_bump" / borsh.U8,
    "arms_bump" / borsh.U8,
    "escrow_auth_bump" / borsh.U8,
)


class ProcessCloseAccountsAccounts(typing.TypedDict):
    player_account: PublicKey
    ship_staking_account: PublicKey
    score_vars_account: PublicKey
    ship_token_account_escrow: PublicKey
    fuel_token_account_escrow: PublicKey
    food_token_account_escrow: PublicKey
    arms_token_account_escrow: PublicKey
    escrow_authority: PublicKey
    token_program: PublicKey
    system_program: PublicKey
    ship_mint: PublicKey
    fuel_mint: PublicKey
    food_mint: PublicKey
    arms_mint: PublicKey
    clock: PublicKey


def process_close_accounts(
    args: ProcessCloseAccountsArgs, accounts: ProcessCloseAccountsAccounts
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["player_account"], is_signer=True, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["ship_staking_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["score_vars_account"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["ship_token_account_escrow"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["fuel_token_account_escrow"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["food_token_account_escrow"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["arms_token_account_escrow"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["escrow_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["token_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["ship_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["fuel_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["food_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["arms_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["clock"], is_signer=False, is_writable=False),
    ]
    identifier = b"\xac\x02\xc5\xaaq\xe1\xbeZ"
    encoded_args = layout.build(
        {
            "staking_bump": args["staking_bump"],
            "scorevars_bump": args["scorevars_bump"],
            "ship_bump": args["ship_bump"],
            "fuel_bump": args["fuel_bump"],
            "food_bump": args["food_bump"],
            "arms_bump": args["arms_bump"],
            "escrow_auth_bump": args["escrow_auth_bump"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



import typing

from base64 import b64decode
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment

from anchorpy.utils.rpc import get_multiple_accounts

from .marketplace.accounts import MarketVars
from .marketplace.accounts import OpenOrdersCounter
from .marketplace.accounts import OrderAccount
from .marketplace.accounts import RegisteredCurrency
from .score.accounts import ScoreVars
from .score.accounts import ScoreVarsShip
from .score.accounts import ShipStaking
from .faction.accounts import PlayerFactionData
from .faction.program_id import PROGRAM_ID as faction_PROGRAM_ID
from .marketplace.program_id import PROGRAM_ID as marketplace_PROGRAM_ID
from .score.program_id import PROGRAM_ID as score_PROGRAM_ID

account_type_map = {}

for account_type_meta in [
        (marketplace_PROGRAM_ID, MarketVars),
        (marketplace_PROGRAM_ID, OpenOrdersCounter),
        (marketplace_PROGRAM_ID, OrderAccount),
        (marketplace_PROGRAM_ID, RegisteredCurrency),
        (score_PROGRAM_ID, ScoreVars),
        (score_PROGRAM_ID, ScoreVarsShip),
        (score_PROGRAM_ID,ShipStaking),
        (faction_PROGRAM_ID,PlayerFactionData)
    ]:
    program = account_type_map.get(account_type_meta[0], {})
    program[account_type_meta[1].discriminator] = account_type_meta[1]
    account_type_map[account_type_meta[0]] = program

async def fetch_multiple_accounts(
        client:AsyncClient,
        addresses: list[PublicKey],
        commitment: Commitment = Commitment('confirmed')
    ):
    resps = await get_multiple_accounts(client, addresses, commitment=commitment)
    result = []
    for resp in resps:
        if resp is None:
            result.append(None)
            continue
        if resp.account.owner in account_type_map.keys():
            program = account_type_map.get(resp.account.owner)
            if resp.account.data[:8] in program.keys():
                result.append(program.get(resp.account.data[:8]).decode(resp.account.data))
                continue
        result.append(resp)
    return result


from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class ProcessRegisterShipArgs(typing.TypedDict):
    scorevars_ship_bump: int
    scorevars_bump: int
    reward_rate_per_second: int
    fuel_max_reserve: int
    food_max_reserve: int
    arms_max_reserve: int
    toolkit_max_reserve: int
    milliseconds_to_burn_one_fuel: int
    milliseconds_to_burn_one_food: int
    milliseconds_to_burn_one_arms: int
    milliseconds_to_burn_one_toolkit: int


layout = borsh.CStruct(
    "scorevars_ship_bump" / borsh.U8,
    "scorevars_bump" / borsh.U8,
    "reward_rate_per_second" / borsh.U64,
    "fuel_max_reserve" / borsh.U32,
    "food_max_reserve" / borsh.U32,
    "arms_max_reserve" / borsh.U32,
    "toolkit_max_reserve" / borsh.U32,
    "milliseconds_to_burn_one_fuel" / borsh.U32,
    "milliseconds_to_burn_one_food" / borsh.U32,
    "milliseconds_to_burn_one_arms" / borsh.U32,
    "milliseconds_to_burn_one_toolkit" / borsh.U32,
)


class ProcessRegisterShipAccounts(typing.TypedDict):
    update_authority_account: PublicKey
    score_vars_account: PublicKey
    score_vars_ship_account: PublicKey
    ship_mint: PublicKey
    system_program: PublicKey


def process_register_ship(
    args: ProcessRegisterShipArgs, accounts: ProcessRegisterShipAccounts
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["update_authority_account"],
            is_signer=True,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["score_vars_account"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["score_vars_ship_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["ship_mint"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
    ]
    identifier = b"\xc1\x98%T\x080\x0fZ"
    encoded_args = layout.build(
        {
            "scorevars_ship_bump": args["scorevars_ship_bump"],
            "scorevars_bump": args["scorevars_bump"],
            "reward_rate_per_second": args["reward_rate_per_second"],
            "fuel_max_reserve": args["fuel_max_reserve"],
            "food_max_reserve": args["food_max_reserve"],
            "arms_max_reserve": args["arms_max_reserve"],
            "toolkit_max_reserve": args["toolkit_max_reserve"],
            "milliseconds_to_burn_one_fuel": args["milliseconds_to_burn_one_fuel"],
            "milliseconds_to_burn_one_food": args["milliseconds_to_burn_one_food"],
            "milliseconds_to_burn_one_arms": args["milliseconds_to_burn_one_arms"],
            "milliseconds_to_burn_one_toolkit": args[
                "milliseconds_to_burn_one_toolkit"
            ],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



import typing
from anchorpy.error import ProgramError


class FactionTypeError(ProgramError):
    def __init__(self) -> None:
        super().__init__(300, "Faction ID must be 0, 1, or 2.")

    code = 300
    name = "FactionTypeError"
    msg = "Faction ID must be 0, 1, or 2."


CustomError = typing.Union[FactionTypeError]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    300: FactionTypeError(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



import typing
from dataclasses import dataclass
from base64 import b64decode
from construct import Construct
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID
from ..types import royalty_tier

class RegisteredCurrencyJSON(typing.TypedDict):
    token_mint: str
    sa_currency_vault: str
    royalty: int
    bump: int
    royalty_tiers: typing.Optional[list[royalty_tier.RoyaltyTierJSON]]


@dataclass
class RegisteredCurrency:
    discriminator: typing.ClassVar = b"<r\xf4\x86\x10\xa63\x95"
    layout: typing.ClassVar = borsh.CStruct(
        "token_mint" / BorshPubkey,
        "sa_currency_vault" / BorshPubkey,
        "royalty" / borsh.U64,
        "bump" / borsh.U8,
        "royalty_tiers"
        / borsh.Option(
            borsh.Vec(typing.cast(Construct, royalty_tier.RoyaltyTier.layout))
        ),
    )
    token_mint: PublicKey
    sa_currency_vault: PublicKey
    royalty: int
    bump: int
    royalty_tiers: typing.Optional[list[royalty_tier.RoyaltyTier]]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.Optional["RegisteredCurrency"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp["result"]["value"]
        if info is None:
            return None
        if info["owner"] != str(PROGRAM_ID):
            raise ValueError("Account does not belong to this program")
        bytes_data = b64decode(info["data"][0])
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.List[typing.Optional["RegisteredCurrency"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["RegisteredCurrency"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != PROGRAM_ID:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "RegisteredCurrency":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = RegisteredCurrency.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            token_mint=dec.token_mint,
            sa_currency_vault=dec.sa_currency_vault,
            royalty=dec.royalty,
            bump=dec.bump,
            royalty_tiers=(
                None
                if dec.royalty_tiers is None
                else list(
                    map(
                        lambda item: royalty_tier.RoyaltyTier.from_decoded(item),
                        dec.royalty_tiers,
                    )
                )
            ),
        )

    def to_json(self) -> RegisteredCurrencyJSON:
        return {
            "token_mint": str(self.token_mint),
            "sa_currency_vault": str(self.sa_currency_vault),
            "royalty": self.royalty,
            "bump": self.bump,
            "royalty_tiers": (
                None
                if self.royalty_tiers is None
                else list(map(lambda item: item.to_json(), self.royalty_tiers))
            ),
        }

    @classmethod
    def from_json(cls, obj: RegisteredCurrencyJSON) -> "RegisteredCurrency":
        return cls(
            token_mint=PublicKey(obj["token_mint"]),
            sa_currency_vault=PublicKey(obj["sa_currency_vault"]),
            royalty=obj["royalty"],
            bump=obj["bump"],
            royalty_tiers=(
                None
                if obj["royalty_tiers"] is None
                else list(
                    map(
                        lambda item: royalty_tier.RoyaltyTier.from_json(item),
                        obj["royalty_tiers"],
                    )
                )
            ),
        )



import typing
from anchorpy.error import ProgramError


class InvalidDestinationAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "Invalid Destination Token Account")

    code = 6000
    name = "InvalidDestinationAccount"
    msg = "Invalid Destination Token Account"


class InvalidInstruction(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "Invalid instruction.")

    code = 6001
    name = "InvalidInstruction"
    msg = "Invalid instruction."


class InvalidMint(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, "Invalid SPL Token mint")

    code = 6002
    name = "InvalidMint"
    msg = "Invalid SPL Token mint"


class InvalidOfferAccountOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, "Invalid Offer Account Owner")

    code = 6003
    name = "InvalidOfferAccountOwner"
    msg = "Invalid Offer Account Owner"


class InvalidTokenAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, "Invalid SPL Token account")

    code = 6004
    name = "InvalidTokenAccount"
    msg = "Invalid SPL Token account"


class NumericalOverflowError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6005, "Numerical overflow error")

    code = 6005
    name = "NumericalOverflowError"
    msg = "Numerical overflow error"


class InvalidUpdateAuthorityAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6006, "Invalid Update Authority account")

    code = 6006
    name = "InvalidUpdateAuthorityAccount"
    msg = "Invalid Update Authority account"


class InvalidOrderVaultAuthorityAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6007, "Invalid Order Vault Authority account")

    code = 6007
    name = "InvalidOrderVaultAuthorityAccount"
    msg = "Invalid Order Vault Authority account"


class UninitializedTokenAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6008, "Uninitialized Token Account")

    code = 6008
    name = "UninitializedTokenAccount"
    msg = "Uninitialized Token Account"


class InsufficientBalance(ProgramError):
    def __init__(self) -> None:
        super().__init__(6009, "Insufficient Balance")

    code = 6009
    name = "InsufficientBalance"
    msg = "Insufficient Balance"


class InvalidOrderDuration(ProgramError):
    def __init__(self) -> None:
        super().__init__(6010, "Invalid Order Duration")

    code = 6010
    name = "InvalidOrderDuration"
    msg = "Invalid Order Duration"


class InvalidOriginationQty(ProgramError):
    def __init__(self) -> None:
        super().__init__(6011, "Origination quantity must be greater than 0")

    code = 6011
    name = "InvalidOriginationQty"
    msg = "Origination quantity must be greater than 0"


class InsufficientOrderQty(ProgramError):
    def __init__(self) -> None:
        super().__init__(6012, "Insufficient Order Quantity Remaining")

    code = 6012
    name = "InsufficientOrderQty"
    msg = "Insufficient Order Quantity Remaining"


class InvalidRoyalty(ProgramError):
    def __init__(self) -> None:
        super().__init__(6013, "Invalid Royalty Value")

    code = 6013
    name = "InvalidRoyalty"
    msg = "Invalid Royalty Value"


class InvalidCounter(ProgramError):
    def __init__(self) -> None:
        super().__init__(6014, "Invalid Open Order Counter")

    code = 6014
    name = "InvalidCounter"
    msg = "Invalid Open Order Counter"


class MintDecimalError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6015, "Mint must be zero decimal")

    code = 6015
    name = "MintDecimalError"
    msg = "Mint must be zero decimal"


class InvalidOrderAccountError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6016, "Order Account does not match provided account")

    code = 6016
    name = "InvalidOrderAccountError"
    msg = "Order Account does not match provided account"


class InvalidRoyaltyTier(ProgramError):
    def __init__(self) -> None:
        super().__init__(6017, "No royalty tier exists with provided stake amount")

    code = 6017
    name = "InvalidRoyaltyTier"
    msg = "No royalty tier exists with provided stake amount"


class RoyaltyTierLength(ProgramError):
    def __init__(self) -> None:
        super().__init__(6018, "Royalty Tier vector cannot hold any additional tiers")

    code = 6018
    name = "RoyaltyTierLength"
    msg = "Royalty Tier vector cannot hold any additional tiers"


CustomError = typing.Union[
    InvalidDestinationAccount,
    InvalidInstruction,
    InvalidMint,
    InvalidOfferAccountOwner,
    InvalidTokenAccount,
    NumericalOverflowError,
    InvalidUpdateAuthorityAccount,
    InvalidOrderVaultAuthorityAccount,
    UninitializedTokenAccount,
    InsufficientBalance,
    InvalidOrderDuration,
    InvalidOriginationQty,
    InsufficientOrderQty,
    InvalidRoyalty,
    InvalidCounter,
    MintDecimalError,
    InvalidOrderAccountError,
    InvalidRoyaltyTier,
    RoyaltyTierLength,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: InvalidDestinationAccount(),
    6001: InvalidInstruction(),
    6002: InvalidMint(),
    6003: InvalidOfferAccountOwner(),
    6004: InvalidTokenAccount(),
    6005: NumericalOverflowError(),
    6006: InvalidUpdateAuthorityAccount(),
    6007: InvalidOrderVaultAuthorityAccount(),
    6008: UninitializedTokenAccount(),
    6009: InsufficientBalance(),
    6010: InvalidOrderDuration(),
    6011: InvalidOriginationQty(),
    6012: InsufficientOrderQty(),
    6013: InvalidRoyalty(),
    6014: InvalidCounter(),
    6015: MintDecimalError(),
    6016: InvalidOrderAccountError(),
    6017: InvalidRoyaltyTier(),
    6018: RoyaltyTierLength(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



from math import floor

# thanks to
# https://palmo.xyz/post/20200218-human-readable-time-intervals-python/

def time_breakdown(ms):
 """Converts an integer representing number of milliseconds into a dictionary
 representing days, hours, minutes, seconds and milliseconds. Output is
 rounded down to the nearest whole number.

 Parameters:
  ms (integer): number of milliseconds.

 Returns:
  (dictionary): breakdown of total days, hours, minutes, seconds and
  milliseconds.

 Example usage:
  >>> time_breakdown = time_breakdown(123456789)
  {'day': 1, 'hr': 10, 'min': 17, 'sec': 36, 'ms': 789}
 """

 ms_out = ms % 1000;
 sec = floor((ms % (1000 * 60)) / 1000)
 min = floor((ms % (1000 * 60 * 60)) / 1000 / 60)
 hr = floor((ms % (1000 * 60 * 60 * 24)) / 1000 / 60 / 60)
 day = floor(
  (ms % (1000 * 60 * 60 * 24 * 365)) / 1000 / 60 / 60 / 24
 )

 return {
  "day": day,
  "hr": hr,
  "min": min,
  "sec": sec,
  "ms": round(ms_out)
 }

def time_breakdown_string(ms, granularity=5):
 """Converts an integer representing number of milliseconds into a string that
 uses natural language to represent the time quantity.

 Parameters:
  ms (integer): number of milliseconds.
  granularity (integer): the level of detail required.

 Returns:
  (string): string that uses natural language to represent the time quantity.

 Example usage:
  >>> time_breakdown_string(123456789)

  '1 day 10 hours 17 minutes 36 seconds 789 milliseconds'

  >>>time_breakdown_string(
  123456789,
  granularity=3)

  '1 day 10 hours 17 minutes'
 """

 time = time_breakdown(ms)
 time_string = ""

 if time["day"] != 0:
  time_string += str(time["day"]) + "d"

 if time["hr"] != 0:
  time_string += " " + str(time["hr"]) + "h"

 if time["min"] != 0:
  time_string += " " + str(time["min"]) + "m"

 if time["sec"] != 0:
  time_string += " " + str(time["sec"]) + "s"

 #filter to specified granularity
 time_string = time_string.strip().split(" ")[:2*granularity]

 return " ".join(time_string)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class ProcessWithdrawShipsArgs(typing.TypedDict):
    staking_bump: int
    scorevars_bump: int
    scorevars_ship_bump: int
    escrow_auth_bump: int
    escrow_bump: int
    treasury_auth_bump: int
    treasury_bump: int


layout = borsh.CStruct(
    "staking_bump" / borsh.U8,
    "scorevars_bump" / borsh.U8,
    "scorevars_ship_bump" / borsh.U8,
    "escrow_auth_bump" / borsh.U8,
    "escrow_bump" / borsh.U8,
    "treasury_auth_bump" / borsh.U8,
    "treasury_bump" / borsh.U8,
)


class ProcessWithdrawShipsAccounts(typing.TypedDict):
    player_account: PublicKey
    ship_staking_account: PublicKey
    score_vars_account: PublicKey
    score_vars_ship_account: PublicKey
    player_atlas_token_account: PublicKey
    ship_token_account_escrow: PublicKey
    ship_token_account_return: PublicKey
    toolkit_token_account_source: PublicKey
    treasury_token_account: PublicKey
    treasury_authority_account: PublicKey
    escrow_authority: PublicKey
    token_program: PublicKey
    system_program: PublicKey
    clock: PublicKey
    ship_mint: PublicKey
    toolkit_mint: PublicKey


def process_withdraw_ships(
    args: ProcessWithdrawShipsArgs, accounts: ProcessWithdrawShipsAccounts
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["player_account"], is_signer=True, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["ship_staking_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["score_vars_account"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["score_vars_ship_account"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["player_atlas_token_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["ship_token_account_escrow"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["ship_token_account_return"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["toolkit_token_account_source"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["treasury_token_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["treasury_authority_account"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["escrow_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["token_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["clock"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["ship_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["toolkit_mint"], is_signer=False, is_writable=True),
    ]
    identifier = b"K\xc6\x17\xd1\xab\xe4\xcfU"
    encoded_args = layout.build(
        {
            "staking_bump": args["staking_bump"],
            "scorevars_bump": args["scorevars_bump"],
            "scorevars_ship_bump": args["scorevars_ship_bump"],
            "escrow_auth_bump": args["escrow_auth_bump"],
            "escrow_bump": args["escrow_bump"],
            "treasury_auth_bump": args["treasury_auth_bump"],
            "treasury_bump": args["treasury_bump"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



import typing
from dataclasses import dataclass
from base64 import b64decode
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class ShipStakingJSON(typing.TypedDict):
    owner: str
    faction_id: int
    ship_mint: str
    ship_quantity_in_escrow: int
    fuel_quantity_in_escrow: int
    food_quantity_in_escrow: int
    arms_quantity_in_escrow: int
    fuel_current_capacity: int
    food_current_capacity: int
    arms_current_capacity: int
    health_current_capacity: int
    staked_at_timestamp: int
    fueled_at_timestamp: int
    fed_at_timestamp: int
    armed_at_timestamp: int
    repaired_at_timestamp: int
    current_capacity_timestamp: int
    total_time_staked: int
    staked_time_paid: int
    pending_rewards: int
    total_rewards_paid: int


@dataclass
class ShipStaking:
    discriminator: typing.ClassVar = b"C\xe5T(\xbc%\t{"
    layout: typing.ClassVar = borsh.CStruct(
        "owner" / BorshPubkey,
        "faction_id" / borsh.U8,
        "ship_mint" / BorshPubkey,
        "ship_quantity_in_escrow" / borsh.U64,
        "fuel_quantity_in_escrow" / borsh.U64,
        "food_quantity_in_escrow" / borsh.U64,
        "arms_quantity_in_escrow" / borsh.U64,
        "fuel_current_capacity" / borsh.U64,
        "food_current_capacity" / borsh.U64,
        "arms_current_capacity" / borsh.U64,
        "health_current_capacity" / borsh.U64,
        "staked_at_timestamp" / borsh.I64,
        "fueled_at_timestamp" / borsh.I64,
        "fed_at_timestamp" / borsh.I64,
        "armed_at_timestamp" / borsh.I64,
        "repaired_at_timestamp" / borsh.I64,
        "current_capacity_timestamp" / borsh.I64,
        "total_time_staked" / borsh.U64,
        "staked_time_paid" / borsh.U64,
        "pending_rewards" / borsh.U64,
        "total_rewards_paid" / borsh.U64,
    )
    owner: PublicKey
    faction_id: int
    ship_mint: PublicKey
    ship_quantity_in_escrow: int
    fuel_quantity_in_escrow: int
    food_quantity_in_escrow: int
    arms_quantity_in_escrow: int
    fuel_current_capacity: int
    food_current_capacity: int
    arms_current_capacity: int
    health_current_capacity: int
    staked_at_timestamp: int
    fueled_at_timestamp: int
    fed_at_timestamp: int
    armed_at_timestamp: int
    repaired_at_timestamp: int
    current_capacity_timestamp: int
    total_time_staked: int
    staked_time_paid: int
    pending_rewards: int
    total_rewards_paid: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.Optional["ShipStaking"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp["result"]["value"]
        if info is None:
            return None
        if info["owner"] != str(PROGRAM_ID):
            raise ValueError("Account does not belong to this program")
        bytes_data = b64decode(info["data"][0])
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.List[typing.Optional["ShipStaking"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["ShipStaking"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != PROGRAM_ID:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "ShipStaking":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = ShipStaking.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            owner=dec.owner,
            faction_id=dec.faction_id,
            ship_mint=dec.ship_mint,
            ship_quantity_in_escrow=dec.ship_quantity_in_escrow,
            fuel_quantity_in_escrow=dec.fuel_quantity_in_escrow,
            food_quantity_in_escrow=dec.food_quantity_in_escrow,
            arms_quantity_in_escrow=dec.arms_quantity_in_escrow,
            fuel_current_capacity=dec.fuel_current_capacity,
            food_current_capacity=dec.food_current_capacity,
            arms_current_capacity=dec.arms_current_capacity,
            health_current_capacity=dec.health_current_capacity,
            staked_at_timestamp=dec.staked_at_timestamp,
            fueled_at_timestamp=dec.fueled_at_timestamp,
            fed_at_timestamp=dec.fed_at_timestamp,
            armed_at_timestamp=dec.armed_at_timestamp,
            repaired_at_timestamp=dec.repaired_at_timestamp,
            current_capacity_timestamp=dec.current_capacity_timestamp,
            total_time_staked=dec.total_time_staked,
            staked_time_paid=dec.staked_time_paid,
            pending_rewards=dec.pending_rewards,
            total_rewards_paid=dec.total_rewards_paid,
        )

    def to_json(self) -> ShipStakingJSON:
        return {
            "owner": str(self.owner),
            "faction_id": self.faction_id,
            "ship_mint": str(self.ship_mint),
            "ship_quantity_in_escrow": self.ship_quantity_in_escrow,
            "fuel_quantity_in_escrow": self.fuel_quantity_in_escrow,
            "food_quantity_in_escrow": self.food_quantity_in_escrow,
            "arms_quantity_in_escrow": self.arms_quantity_in_escrow,
            "fuel_current_capacity": self.fuel_current_capacity,
            "food_current_capacity": self.food_current_capacity,
            "arms_current_capacity": self.arms_current_capacity,
            "health_current_capacity": self.health_current_capacity,
            "staked_at_timestamp": self.staked_at_timestamp,
            "fueled_at_timestamp": self.fueled_at_timestamp,
            "fed_at_timestamp": self.fed_at_timestamp,
            "armed_at_timestamp": self.armed_at_timestamp,
            "repaired_at_timestamp": self.repaired_at_timestamp,
            "current_capacity_timestamp": self.current_capacity_timestamp,
            "total_time_staked": self.total_time_staked,
            "staked_time_paid": self.staked_time_paid,
            "pending_rewards": self.pending_rewards,
            "total_rewards_paid": self.total_rewards_paid,
        }

    @classmethod
    def from_json(cls, obj: ShipStakingJSON) -> "ShipStaking":
        return cls(
            owner=PublicKey(obj["owner"]),
            faction_id=obj["faction_id"],
            ship_mint=PublicKey(obj["ship_mint"]),
            ship_quantity_in_escrow=obj["ship_quantity_in_escrow"],
            fuel_quantity_in_escrow=obj["fuel_quantity_in_escrow"],
            food_quantity_in_escrow=obj["food_quantity_in_escrow"],
            arms_quantity_in_escrow=obj["arms_quantity_in_escrow"],
            fuel_current_capacity=obj["fuel_current_capacity"],
            food_current_capacity=obj["food_current_capacity"],
            arms_current_capacity=obj["arms_current_capacity"],
            health_current_capacity=obj["health_current_capacity"],
            staked_at_timestamp=obj["staked_at_timestamp"],
            fueled_at_timestamp=obj["fueled_at_timestamp"],
            fed_at_timestamp=obj["fed_at_timestamp"],
            armed_at_timestamp=obj["armed_at_timestamp"],
            repaired_at_timestamp=obj["repaired_at_timestamp"],
            current_capacity_timestamp=obj["current_capacity_timestamp"],
            total_time_staked=obj["total_time_staked"],
            staked_time_paid=obj["staked_time_paid"],
            pending_rewards=obj["pending_rewards"],
            total_rewards_paid=obj["total_rewards_paid"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class ArmsJSON(typing.TypedDict):
    kind: typing.Literal["Arms"]


class FuelJSON(typing.TypedDict):
    kind: typing.Literal["Fuel"]


class FoodJSON(typing.TypedDict):
    kind: typing.Literal["Food"]


class ToolkitsJSON(typing.TypedDict):
    kind: typing.Literal["Toolkits"]


@dataclass
class Arms:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Arms"

    @classmethod
    def to_json(cls) -> ArmsJSON:
        return ArmsJSON(
            kind="Arms",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Arms": {},
        }


@dataclass
class Fuel:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Fuel"

    @classmethod
    def to_json(cls) -> FuelJSON:
        return FuelJSON(
            kind="Fuel",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Fuel": {},
        }


@dataclass
class Food:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Food"

    @classmethod
    def to_json(cls) -> FoodJSON:
        return FoodJSON(
            kind="Food",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Food": {},
        }


@dataclass
class Toolkits:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "Toolkits"

    @classmethod
    def to_json(cls) -> ToolkitsJSON:
        return ToolkitsJSON(
            kind="Toolkits",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Toolkits": {},
        }


ResourceKind = typing.Union[Arms, Fuel, Food, Toolkits]
ResourceJSON = typing.Union[ArmsJSON, FuelJSON, FoodJSON, ToolkitsJSON]


def from_decoded(obj: dict) -> ResourceKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Arms" in obj:
        return Arms()
    if "Fuel" in obj:
        return Fuel()
    if "Food" in obj:
        return Food()
    if "Toolkits" in obj:
        return Toolkits()
    raise ValueError("Invalid enum object")


def from_json(obj: ResourceJSON) -> ResourceKind:
    if obj["kind"] == "Arms":
        return Arms()
    if obj["kind"] == "Fuel":
        return Fuel()
    if obj["kind"] == "Food":
        return Food()
    if obj["kind"] == "Toolkits":
        return Toolkits()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Arms" / borsh.CStruct(),
    "Fuel" / borsh.CStruct(),
    "Food" / borsh.CStruct(),
    "Toolkits" / borsh.CStruct(),
)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class ProcessInitializeArgs(typing.TypedDict):
    bump: int
    treasury_bump: int
    treasury_auth_bump: int


layout = borsh.CStruct(
    "bump" / borsh.U8, "treasury_bump" / borsh.U8, "treasury_auth_bump" / borsh.U8
)


class ProcessInitializeAccounts(typing.TypedDict):
    update_authority_account: PublicKey
    score_vars_account: PublicKey
    token_program: PublicKey
    system_program: PublicKey
    rent: PublicKey
    treasury_token_account: PublicKey
    treasury_authority_account: PublicKey
    atlas_mint: PublicKey
    fuel_mint: PublicKey
    food_mint: PublicKey
    arms_mint: PublicKey
    toolkit_mint: PublicKey


def process_initialize(
    args: ProcessInitializeArgs, accounts: ProcessInitializeAccounts
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["update_authority_account"],
            is_signer=True,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["score_vars_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["token_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["rent"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["treasury_token_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["treasury_authority_account"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(pubkey=accounts["atlas_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["fuel_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["food_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["arms_mint"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["toolkit_mint"], is_signer=False, is_writable=False
        ),
    ]
    identifier = b"\xb8X\x14\xad\x113Q\x85"
    encoded_args = layout.build(
        {
            "bump": args["bump"],
            "treasury_bump": args["treasury_bump"],
            "treasury_auth_bump": args["treasury_auth_bump"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



import asyncio, json, httpx, prettytable
import asyncclick as click
from solana.publickey import PublicKey
from typing import Any, List, Tuple, Union
from .score import  getShipStakingAccount, getScoreVarsShipAccount, getScoreEscrowAuthAccount
from solana.rpc.async_api import AsyncClient
from . import fetch_multiple_accounts
from .score import ScoreVars, ShipStaking, ScoreStats
from .utils.semantic_time import time_breakdown_string

SOLANA_NODE_URL = 'https://api.mainnet-beta.solana.com'
NFTS_URL = 'https://api.staratlas.club/nfts?category=ship'
ASSET_PRICES_URL = 'https://api.staratlas.club/redis/asset-prices?symbols='

@click.group()
async def cli():
    """
    Main cli group
    """

@cli.command()
@click.option("--currency", "-c", show_default=True, default='ATLAS', type=click.Choice(['ATLAS', 'USDC']), help="Currency, either ATLAS or USDC")
@click.argument('wallet', nargs=1)
async def score_fleet(wallet, currency):
    """
    Get score fleet info
    """
    def get_price(data, sym, quote):
        markets = {d.get('quote_symbol'):d for d in data.get(sym,{}).get('data',[])}
        atlas_usdc = data.get(sym,{}).get("ATLAS/USDC",1)
        price =  markets.get(quote, {}).get('market_price',0)
        if not price:
            if quote == 'ATLAS':
                price = markets.get('USDC', {}).get('market_price',0) / atlas_usdc
            elif quote == 'USDC':
                price = markets.get('ATLAS', {}).get('market_price',0) * atlas_usdc
            if not price and sym == 'ATLAS':
                price = atlas_usdc
        return price
    playerKey = PublicKey(wallet)
    connection = AsyncClient(SOLANA_NODE_URL)
    async with httpx.AsyncClient() as client:
        r = await client.get(NFTS_URL)
        nfts = r.json()
        await client.aclose()
    async with httpx.AsyncClient() as client:
        r = await client.get(ASSET_PRICES_URL + 'ATLAS,AMMO,FUEL,FOOD,TOOL,'+\
                             ','.join([nft.get('symbol') for nft in nfts]))
        prices = r.json()
        await client.aclose()
    mints = [PublicKey(nft.get('mint')) for nft in nfts]
    staking = [getShipStakingAccount(playerKey, mint)[0] for mint in mints]
    escrow = [getScoreEscrowAuthAccount(playerKey, mint)[0] for mint in mints]
    vars = [getScoreVarsShipAccount(mint)[0] for mint in mints]
    staking_state = await fetch_multiple_accounts(connection, staking)
    vars_state = await fetch_multiple_accounts(connection, vars)
    await connection.close()
    pt = prettytable.PrettyTable()
    #pt.field_names = ['Ship', 'Qty', 'Value', 'Daily Rewards', 'Daily Burn', 'Net Yield', 'APR']
    pt.field_names = ['Ship', 'Qty', 'Daily Rewards', 'Daily Burn', 'Net Yield']
    pt.align['Ship'] = 'l'
    pt.align['Qty'] = 'r'
    pt.align['Value'] = 'r'
    pt.align['Daily Rewards'] = 'r'
    pt.align['Daily Burn'] = 'r'
    pt.align['Net Yield'] = 'r'
    pt.align['APR'] = 'r'
    for i in range(len(nfts)):
        if staking_state[i]:
            qty = staking_state[i].ship_quantity_in_escrow
            value = get_price(prices, nfts[i].get('symbol'), currency) * qty
            reward = (60 * 60 * 24 * vars_state[i].reward_rate_per_second * \
                 staking_state[i].ship_quantity_in_escrow) / 10 ** 8
            if currency == 'USDC':
                reward = reward * get_price(prices, 'ATLAS','USDC')
            stats = ScoreStats(vars_state[i], staking_state[i])
            burn = stats.arms_daily_burn * get_price(prices, 'AMMO', currency) + \
                   stats.food_daily_burn * get_price(prices, 'FOOD', currency) + \
                   stats.fuel_daily_burn * get_price(prices, 'FUEL', currency) + \
                   stats.toolkit_daily_burn * get_price(prices, 'TOOL', currency)
            net = reward - burn
            if value:
                apr = 100*net * 365 / value
            else:
                apr = 0
            pt.add_row([
                nfts[i].get('name'),
                '{:,.0f}'.format(qty),
                #'{:,.0f}'.format(value),
                '{:,.2f}'.format(reward),
                '{:,.2f}'.format(burn),
                '{:,.2f}'.format(net)
                #'{:,.1f} %'.format(apr)
            ])
    print(pt)


@cli.command()
@click.argument('wallet', nargs=1)
async def score_supplies(wallet):
    """
    Get score supplies state
    """
    playerKey = PublicKey(wallet)
    connection = AsyncClient(SOLANA_NODE_URL)
    async with httpx.AsyncClient() as client:
        r = await client.get(NFTS_URL)
        nfts = r.json()
        await client.aclose()
    mints = [PublicKey(nft.get('mint')) for nft in nfts]
    staking = [getShipStakingAccount(playerKey, mint)[0] for mint in mints]
    escrow = [getScoreEscrowAuthAccount(playerKey, mint)[0] for mint in mints]
    vars = [getScoreVarsShipAccount(mint)[0] for mint in mints]
    staking_state = await fetch_multiple_accounts(connection, staking)
    vars_state = await fetch_multiple_accounts(connection, vars)
    await connection.close()
    pt = prettytable.PrettyTable()
    pt.field_names = ['Ship', 'Qty', 'Fuel', 'Food', 'Ammo', 'Tool', 'Resupply in']
    pt.align['Ship'] = 'l'
    pt.align['Qty'] = 'r'
    pt.align['Fuel'] = 'r'
    pt.align['Food'] = 'r'
    pt.align['Ammo'] = 'r'
    pt.align['Tool'] = 'r'
    pt.align['Resupply in'] = 'l'
    for i in range(len(nfts)):
        if staking_state[i]:
            qty = staking_state[i].ship_quantity_in_escrow
            reward = (60 * 60 * 24 * vars_state[i].reward_rate_per_second * \
                 staking_state[i].ship_quantity_in_escrow) / 10 ** 8
            stats = ScoreStats(vars_state[i], staking_state[i])
            pt.add_row([
                nfts[i].get('name'),
                '{:,.0f}'.format(qty),
                '{:,.1f}%'.format(stats.fuel_current_supply_to_total_capacity_percent*100),
                '{:,.1f}%'.format(stats.food_current_supply_to_total_capacity_percent*100),
                '{:,.1f}%'.format(stats.arms_current_supply_to_total_capacity_percent*100),
                '{:,.1f}%'.format(stats.toolkit_current_supply_to_total_capacity_percent*100),
                time_breakdown_string(stats.seconds_remaining * 1000, 1 )
            ])
    print(pt)


if __name__ == '__main__':
    cli(_anyio_backend="asyncio")


from solana.publickey import PublicKey
from .program_id import PROGRAM_ID
from .accounts import ScoreVars, ShipStaking

import  time, math

def getScoreVarsShipAccount(shipMint: PublicKey, programId: PublicKey = PROGRAM_ID):
    """
    Returns the public key and bump seed for the SCORE variables ship account associated with the provided ship mint.

    :param shipMint: Mint address for the desired ship
    :param programId: Deployed program ID for the SCORE program
    :return: [Escrow account public key, bump seed]
    """
    return PublicKey.find_program_address(
        [b'SCOREVARS_SHIP', bytes(shipMint)],
        PROGRAM_ID)

def getScoreEscrowAuthAccount(playerPublicKey: PublicKey, shipMint: PublicKey, programId: PublicKey = PROGRAM_ID):
    """
     Returns the SCORE escrow authority account

    :param playerPublicKey: Player's public key
    :param shipMint: Mint address for the desired ship
    :param programId: Deployed program ID for the SCORE program
    :return: [Authority account public key, bump seed]
    """
    return PublicKey.find_program_address(
        [b'SCORE_ESCROW_AUTHORITY', bytes(playerPublicKey), bytes(shipMint)],
        PROGRAM_ID)

def getShipStakingAccount(playerPublicKey:PublicKey, assetMint:PublicKey, programId: PublicKey = PROGRAM_ID):
    """
    Returns a user's ship staking account

    :param playerPublicKey: Player's public key
    :param assetMint: Mint address for the desired resource
    :param programId: Deployed program ID for the SCORE program
    :return: [Staking account public key, bump seed]
    """
    return PublicKey.find_program_address(
        [b'SCORE_INFO', bytes(playerPublicKey), bytes(assetMint)],
        PROGRAM_ID)

class ScoreStats():
    seconds_remaining: int
    seconds_since_last_action: int
    fuel_daily_burn: int
    fuel_total_capacity_seconds: int
    fuel_current_supply_to_total_capacity_percent: float
    fuel_total_capacity: int
    fuel_current_supply: int
    fuel_needed_for_full_supply: int
    fuel_needed_for_optimal_supply: int
    arms_daily_burn: int
    arms_total_capacity_seconds: int
    arms_current_supply_to_total_capacity_percent: float
    arms_total_capacity: int
    arms_current_supply: int
    arms_needed_for_full_supply: int
    arms_needed_for_optimal_supply: int
    food_daily_burn: int
    food_total_capacity_seconds: int
    food_current_supply_to_total_capacity_percent: float
    food_total_capacity: int
    food_current_supply: int
    food_needed_for_full_supply: int
    food_needed_for_optimal_supply: int
    toolkit_daily_burn: int
    toolkit_total_capacity_seconds: int
    toolkit_current_supply_to_total_capacity_percent: float
    toolkit_total_capacity: int
    toolkit_current_supply: int
    toolkit_needed_for_full_supply: int
    toolkit_needed_for_optimal_supply: int

    def __init__(self, vars: ScoreVars, staking: ShipStaking):
        self.vars = vars
        self.staking = staking
        self.seconds_since_last_action = int(time.time()) - staking.current_capacity_timestamp
        self.fuel_total_capacity_seconds = int(vars.fuel_max_reserve * vars.milliseconds_to_burn_one_fuel / 1000)
        self.food_total_capacity_seconds = int(vars.food_max_reserve * vars.milliseconds_to_burn_one_food / 1000)
        self.arms_total_capacity_seconds = int(vars.arms_max_reserve * vars.milliseconds_to_burn_one_arms / 1000)
        self.toolkit_total_capacity_seconds = int(vars.toolkit_max_reserve * vars.milliseconds_to_burn_one_toolkit / 1000)
        min_total_capacity_seconds = min(self.fuel_total_capacity_seconds,
                                                             self.food_total_capacity_seconds,
                                                             self.arms_total_capacity_seconds,
                                                             self.toolkit_total_capacity_seconds
                                                             )
        self.seconds_remaining = max(0, min_total_capacity_seconds - self.seconds_since_last_action)
        # fuel
        self.fuel_daily_burn = int(staking.ship_quantity_in_escrow * 24 * 60 * 60 * 1000 /
                                   vars.milliseconds_to_burn_one_fuel)
        self.fuel_total_capacity = int(staking.ship_quantity_in_escrow * self.fuel_total_capacity_seconds /
                                       (vars.milliseconds_to_burn_one_fuel / 1000) )
        self.fuel_current_supply_to_total_capacity_percent = max(0, (staking.fuel_current_capacity -
                                                              self.seconds_since_last_action) / \
                                                             self.fuel_total_capacity_seconds)
        self.fuel_needed_for_full_supply = staking.ship_quantity_in_escrow * \
                                           max(0, (1 - self.fuel_current_supply_to_total_capacity_percent) *
                                               self.fuel_total_capacity_seconds) / \
                                           (vars.milliseconds_to_burn_one_fuel/1000)
        self.fuel_current_supply = self.fuel_total_capacity - self.fuel_needed_for_full_supply
        self.fuel_needed_for_optimal_supply = max(0, int(staking.ship_quantity_in_escrow * min_total_capacity_seconds / \
                                                  (vars.milliseconds_to_burn_one_fuel / 1000) -
                                                  self.fuel_current_supply))
        # arms
        self.arms_daily_burn = int(staking.ship_quantity_in_escrow * 24 * 60 * 60 * 1000 /
                                   vars.milliseconds_to_burn_one_arms)
        self.arms_total_capacity = int(staking.ship_quantity_in_escrow * self.arms_total_capacity_seconds /
                                       (vars.milliseconds_to_burn_one_arms / 1000))
        self.arms_current_supply_to_total_capacity_percent = max(0, (staking.arms_current_capacity -
                                                              self.seconds_since_last_action) / \
                                                             self.arms_total_capacity_seconds)
        self.arms_needed_for_full_supply = staking.ship_quantity_in_escrow * \
                                           max(0, (1 - self.arms_current_supply_to_total_capacity_percent) *
                                               self.arms_total_capacity_seconds) / \
                                           (vars.milliseconds_to_burn_one_arms / 1000)
        self.arms_current_supply = self.arms_total_capacity - self.arms_needed_for_full_supply
        self.arms_needed_for_optimal_supply = max(0, int(staking.ship_quantity_in_escrow * min_total_capacity_seconds / \
                                                         (vars.milliseconds_to_burn_one_arms / 1000) -
                                                         self.arms_current_supply))
        # food
        self.food_daily_burn = int(staking.ship_quantity_in_escrow * 24 * 60 * 60 * 1000 /
                                   vars.milliseconds_to_burn_one_food)
        self.food_total_capacity = int(staking.ship_quantity_in_escrow * self.food_total_capacity_seconds /
                                       (vars.milliseconds_to_burn_one_food / 1000))
        self.food_current_supply_to_total_capacity_percent = max(0, (staking.food_current_capacity -
                                                              self.seconds_since_last_action) / \
                                                             self.food_total_capacity_seconds)
        self.food_needed_for_full_supply = staking.ship_quantity_in_escrow * \
                                           max(0, (1 - self.food_current_supply_to_total_capacity_percent) *
                                               self.food_total_capacity_seconds) / \
                                           (vars.milliseconds_to_burn_one_food / 1000)
        self.food_current_supply = self.food_total_capacity - self.food_needed_for_full_supply
        self.food_needed_for_optimal_supply = max(0, int(staking.ship_quantity_in_escrow * min_total_capacity_seconds / \
                                                         (vars.milliseconds_to_burn_one_food / 1000) -
                                                         self.food_current_supply))
        # toolkit
        self.toolkit_daily_burn = int(staking.ship_quantity_in_escrow * 24 * 60 * 60 * 1000 /
                                      vars.milliseconds_to_burn_one_toolkit)
        self.toolkit_total_capacity = int(staking.ship_quantity_in_escrow * self.toolkit_total_capacity_seconds /
                                          (vars.milliseconds_to_burn_one_toolkit / 1000))
        self.toolkit_current_supply_to_total_capacity_percent = max(0, (staking.health_current_capacity -
                                                                 self.seconds_since_last_action) / \
                                                                self.toolkit_total_capacity_seconds)
        self.toolkit_needed_for_full_supply = staking.ship_quantity_in_escrow * \
                                              max(0, (1 - self.toolkit_current_supply_to_total_capacity_percent) *
                                                  self.toolkit_total_capacity_seconds) / \
                                              (vars.milliseconds_to_burn_one_toolkit / 1000)
        self.toolkit_current_supply = self.toolkit_total_capacity - self.toolkit_needed_for_full_supply
        self.toolkit_needed_for_optimal_supply = max(0,
                                                     int(staking.ship_quantity_in_escrow * min_total_capacity_seconds / \
                                                         (vars.milliseconds_to_burn_one_toolkit / 1000) -
                                                         self.toolkit_current_supply))

    def limited_atlas_resupply(self,
                               atlas: float,
                               fuel_price: float = 0.00144336,
                               food_price: float = 0.0006144,
                               arms_price: float = 0.00215039,
                               toolkit_price: int = 0.0017408) -> list[4]:
        resources = {
            "food": {
                "seconds_to_optimal_supply": self.food_needed_for_optimal_supply * (
                        self.vars.milliseconds_to_burn_one_food / 1000) / self.staking.ship_quantity_in_escrow,
                "atlas_per_second": self.staking.ship_quantity_in_escrow * food_price / (
                        self.vars.milliseconds_to_burn_one_food / 1000),
                "milliseconds_to_burn_one":self.vars.milliseconds_to_burn_one_food
            },
            "fuel": {
                "seconds_to_optimal_supply": self.fuel_needed_for_optimal_supply * (
                        self.vars.milliseconds_to_burn_one_fuel / 1000) / self.staking.ship_quantity_in_escrow,
                "atlas_per_second": self.staking.ship_quantity_in_escrow * fuel_price / (
                        self.vars.milliseconds_to_burn_one_fuel / 1000),
                "milliseconds_to_burn_one":self.vars.milliseconds_to_burn_one_fuel
            },
            "arms": {
                "seconds_to_optimal_supply": self.arms_needed_for_optimal_supply * (
                        self.vars.milliseconds_to_burn_one_arms / 1000) / self.staking.ship_quantity_in_escrow,
                "atlas_per_second": self.staking.ship_quantity_in_escrow * arms_price / (
                        self.vars.milliseconds_to_burn_one_arms / 1000),
                "milliseconds_to_burn_one":self.vars.milliseconds_to_burn_one_arms
            },
            "toolkit": {
                "seconds_to_optimal_supply": self.toolkit_needed_for_optimal_supply * (
                        self.vars.milliseconds_to_burn_one_toolkit / 1000) / self.staking.ship_quantity_in_escrow,
                "atlas_per_second": self.staking.ship_quantity_in_escrow * toolkit_price / (
                        self.vars.milliseconds_to_burn_one_toolkit / 1000),
                "milliseconds_to_burn_one":self.vars.milliseconds_to_burn_one_toolkit
            }
        }

        def get_depleted_resources():
            return [k for k in resources.keys() if not resources[k].get('seconds_to_optimal_supply', 0) > 0]

        def get_atlas_per_second():
            return sum(
                [resources[k]['atlas_per_second'] for k in resources.keys() if k not in get_depleted_resources()])

        def update_consumed_seconds(secs=0):
            for k in resources.keys():
                if k not in get_depleted_resources():
                    resources[k]["seconds_to_optimal_supply"] -= secs
                    resources[k]["seconds_consumed"] = resources[k].get('seconds_consumed', 0) + secs

        atlas_remaining = atlas
        atlas_consumed = None
        for m in sorted([{**{"key": k}, **resources[k]} for k in resources.keys() if k not in get_depleted_resources()],
                        key=lambda x: x.get('seconds_to_optimal_supply')):
            if not get_atlas_per_second() > 0 or atlas_consumed == 0:
                break
            seconds_consumed = max(0, min(atlas_remaining / get_atlas_per_second(),
                                          resources[m.get('key')]['seconds_to_optimal_supply']))
            if not seconds_consumed > 0:
                break
            atlas_consumed = seconds_consumed * get_atlas_per_second()
            atlas_remaining = atlas_remaining - atlas_consumed
            update_consumed_seconds(seconds_consumed)
        return{ k:math.floor(resources[k].get('seconds_consumed',0)/\
                  (resources[k].get('milliseconds_to_burn_one')/1000)) for k in resources}

    def to_json(self):
        return {
            "seconds_remaining": int(self.seconds_remaining),
            "seconds_since_last_action": int(self.seconds_since_last_action),
            "fuel_daily_burn": int(self.fuel_daily_burn),
            "fuel_total_capacity_seconds": int(self.fuel_total_capacity_seconds),
            "fuel_current_supply_to_total_capacity_percent": float(self.fuel_current_supply_to_total_capacity_percent),
            "fuel_total_capacity": int(self.fuel_total_capacity),
            "fuel_current_supply": int(self.fuel_current_supply),
            "fuel_needed_for_full_supply": int(self.fuel_needed_for_full_supply),
            "fuel_needed_for_optimal_supply": int(self.fuel_needed_for_optimal_supply),
            "arms_daily_burn": int(self.arms_daily_burn),
            "arms_total_capacity_seconds": int(self.arms_total_capacity_seconds),
            "arms_current_supply_to_total_capacity_percent": float(self.arms_current_supply_to_total_capacity_percent),
            "arms_total_capacity": int(self.arms_total_capacity),
            "arms_current_supply": int(self.arms_current_supply),
            "arms_needed_for_full_supply": int(self.arms_needed_for_full_supply),
            "arms_needed_for_optimal_supply": int(self.arms_needed_for_optimal_supply),
            "food_daily_burn": int(self.food_daily_burn),
            "food_total_capacity_seconds": int(self.food_total_capacity_seconds),
            "food_current_supply_to_total_capacity_percent": float(self.food_current_supply_to_total_capacity_percent),
            "food_total_capacity": int(self.food_total_capacity),
            "food_current_supply": int(self.food_current_supply),
            "food_needed_for_full_supply": int(self.food_needed_for_full_supply),
            "food_needed_for_optimal_supply": int(self.food_needed_for_optimal_supply),
            "toolkit_daily_burn": int(self.toolkit_daily_burn),
            "toolkit_total_capacity_seconds": int(self.toolkit_total_capacity_seconds),
            "toolkit_current_supply_to_total_capacity_percent": float(
                self.toolkit_current_supply_to_total_capacity_percent),
            "toolkit_total_capacity": int(self.toolkit_total_capacity),
            "toolkit_current_supply": int(self.toolkit_current_supply),
            "toolkit_needed_for_full_supply": int(self.toolkit_needed_for_full_supply),
            "toolkit_needed_for_optimal_supply": int(self.toolkit_needed_for_optimal_supply)
        }



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class InitializeMarketplaceAccounts(typing.TypedDict):
    update_authority_account: PublicKey
    market_vars_account: PublicKey
    system_program: PublicKey


def initialize_marketplace(
    accounts: InitializeMarketplaceAccounts,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["update_authority_account"],
            is_signer=True,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["market_vars_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
    ]
    identifier = b"/Q@\x00`8i\x07"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



import typing
from anchorpy.error import ProgramError


class ScorevarsNotInitialized(ProgramError):
    def __init__(self) -> None:
        super().__init__(300, "Scorevars auth is not initialized")

    code = 300
    name = "ScorevarsNotInitialized"
    msg = "Scorevars auth is not initialized"


class ScorevarsAuthInvalid(ProgramError):
    def __init__(self) -> None:
        super().__init__(301, "Scorevars auth invalid")

    code = 301
    name = "ScorevarsAuthInvalid"
    msg = "Scorevars auth invalid"


class FactionTypeError(ProgramError):
    def __init__(self) -> None:
        super().__init__(302, "Faction ID must be 0, 1, or 2.")

    code = 302
    name = "FactionTypeError"
    msg = "Faction ID must be 0, 1, or 2."


class InvalidShipError(ProgramError):
    def __init__(self) -> None:
        super().__init__(303, "Invalid Ship Mint")

    code = 303
    name = "InvalidShipError"
    msg = "Invalid Ship Mint"


class InvalidResourceError(ProgramError):
    def __init__(self) -> None:
        super().__init__(304, "Invalid Resource Mint")

    code = 304
    name = "InvalidResourceError"
    msg = "Invalid Resource Mint"


class ZeroResourceError(ProgramError):
    def __init__(self) -> None:
        super().__init__(305, "Resource Quantity must be >0.")

    code = 305
    name = "ZeroResourceError"
    msg = "Resource Quantity must be >0."


class ZeroShipError(ProgramError):
    def __init__(self) -> None:
        super().__init__(306, "Ship Quantity must be >0.")

    code = 306
    name = "ZeroShipError"
    msg = "Ship Quantity must be >0."


class NumericalOverflowError(ProgramError):
    def __init__(self) -> None:
        super().__init__(307, "Numerical overflow error")

    code = 307
    name = "NumericalOverflowError"
    msg = "Numerical overflow error"


class ResourceAmountTooSmall(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            308, "Resource Amount Too Small, would add 0 seconds to capacity"
        )

    code = 308
    name = "ResourceAmountTooSmall"
    msg = "Resource Amount Too Small, would add 0 seconds to capacity"


class InvalidScoreVarsAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            309, "Invalid ScoreVars account for ship Staking Account given"
        )

    code = 309
    name = "InvalidScoreVarsAccount"
    msg = "Invalid ScoreVars account for ship Staking Account given"


class InvalidResourceWithdraw(ProgramError):
    def __init__(self) -> None:
        super().__init__(310, "Invalid Resource Withdraw, nothing to withdraw")

    code = 310
    name = "InvalidResourceWithdraw"
    msg = "Invalid Resource Withdraw, nothing to withdraw"


class InvalidShipStakingOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(311, "Invalid Ship Staking Owner")

    code = 311
    name = "InvalidShipStakingOwner"
    msg = "Invalid Ship Staking Owner"


class InvalidShipWithdraw(ProgramError):
    def __init__(self) -> None:
        super().__init__(312, "Invalid Ship Withdraw")

    code = 312
    name = "InvalidShipWithdraw"
    msg = "Invalid Ship Withdraw"


class NotEnoughToolkits(ProgramError):
    def __init__(self) -> None:
        super().__init__(313, "Not enough toolkits for Ship Withdraw")

    code = 313
    name = "NotEnoughToolkits"
    msg = "Not enough toolkits for Ship Withdraw"


class EscrowAccountNotZero(ProgramError):
    def __init__(self) -> None:
        super().__init__(314, "Escrow account must be 0 to close accounts")

    code = 314
    name = "EscrowAccountNotZero"
    msg = "Escrow account must be 0 to close accounts"


class IncorrectTokenAccountOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(315, "Token account does not have correct owner!")

    code = 315
    name = "IncorrectTokenAccountOwner"
    msg = "Token account does not have correct owner!"


class UninitializedTokenAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(316, "Token or Mint Account is not initialized!")

    code = 316
    name = "UninitializedTokenAccount"
    msg = "Token or Mint Account is not initialized!"


class IncorrectTokenAccountMint(ProgramError):
    def __init__(self) -> None:
        super().__init__(317, "Token or Mint Account mint is not correct!")

    code = 317
    name = "IncorrectTokenAccountMint"
    msg = "Token or Mint Account mint is not correct!"


CustomError = typing.Union[
    ScorevarsNotInitialized,
    ScorevarsAuthInvalid,
    FactionTypeError,
    InvalidShipError,
    InvalidResourceError,
    ZeroResourceError,
    ZeroShipError,
    NumericalOverflowError,
    ResourceAmountTooSmall,
    InvalidScoreVarsAccount,
    InvalidResourceWithdraw,
    InvalidShipStakingOwner,
    InvalidShipWithdraw,
    NotEnoughToolkits,
    EscrowAccountNotZero,
    IncorrectTokenAccountOwner,
    UninitializedTokenAccount,
    IncorrectTokenAccountMint,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    300: ScorevarsNotInitialized(),
    301: ScorevarsAuthInvalid(),
    302: FactionTypeError(),
    303: InvalidShipError(),
    304: InvalidResourceError(),
    305: ZeroResourceError(),
    306: ZeroShipError(),
    307: NumericalOverflowError(),
    308: ResourceAmountTooSmall(),
    309: InvalidScoreVarsAccount(),
    310: InvalidResourceWithdraw(),
    311: InvalidShipStakingOwner(),
    312: InvalidShipWithdraw(),
    313: NotEnoughToolkits(),
    314: EscrowAccountNotZero(),
    315: IncorrectTokenAccountOwner(),
    316: UninitializedTokenAccount(),
    317: IncorrectTokenAccountMint(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



import typing
from dataclasses import dataclass
from base64 import b64decode
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class ScoreVarsShipJSON(typing.TypedDict):
    ship_mint: str
    reward_rate_per_second: int
    fuel_max_reserve: int
    food_max_reserve: int
    arms_max_reserve: int
    toolkit_max_reserve: int
    milliseconds_to_burn_one_fuel: int
    milliseconds_to_burn_one_food: int
    milliseconds_to_burn_one_arms: int
    milliseconds_to_burn_one_toolkit: int


@dataclass
class ScoreVarsShip:
    discriminator: typing.ClassVar = b"\x1b\xda\x89\x99j\x86\xda\x8f"
    layout: typing.ClassVar = borsh.CStruct(
        "ship_mint" / BorshPubkey,
        "reward_rate_per_second" / borsh.U64,
        "fuel_max_reserve" / borsh.U32,
        "food_max_reserve" / borsh.U32,
        "arms_max_reserve" / borsh.U32,
        "toolkit_max_reserve" / borsh.U32,
        "milliseconds_to_burn_one_fuel" / borsh.U32,
        "milliseconds_to_burn_one_food" / borsh.U32,
        "milliseconds_to_burn_one_arms" / borsh.U32,
        "milliseconds_to_burn_one_toolkit" / borsh.U32,
    )
    ship_mint: PublicKey
    reward_rate_per_second: int
    fuel_max_reserve: int
    food_max_reserve: int
    arms_max_reserve: int
    toolkit_max_reserve: int
    milliseconds_to_burn_one_fuel: int
    milliseconds_to_burn_one_food: int
    milliseconds_to_burn_one_arms: int
    milliseconds_to_burn_one_toolkit: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.Optional["ScoreVarsShip"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp["result"]["value"]
        if info is None:
            return None
        if info["owner"] != str(PROGRAM_ID):
            raise ValueError("Account does not belong to this program")
        bytes_data = b64decode(info["data"][0])
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.List[typing.Optional["ScoreVarsShip"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["ScoreVarsShip"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != PROGRAM_ID:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "ScoreVarsShip":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = ScoreVarsShip.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            ship_mint=dec.ship_mint,
            reward_rate_per_second=dec.reward_rate_per_second,
            fuel_max_reserve=dec.fuel_max_reserve,
            food_max_reserve=dec.food_max_reserve,
            arms_max_reserve=dec.arms_max_reserve,
            toolkit_max_reserve=dec.toolkit_max_reserve,
            milliseconds_to_burn_one_fuel=dec.milliseconds_to_burn_one_fuel,
            milliseconds_to_burn_one_food=dec.milliseconds_to_burn_one_food,
            milliseconds_to_burn_one_arms=dec.milliseconds_to_burn_one_arms,
            milliseconds_to_burn_one_toolkit=dec.milliseconds_to_burn_one_toolkit,
        )

    def to_json(self) -> ScoreVarsShipJSON:
        return {
            "ship_mint": str(self.ship_mint),
            "reward_rate_per_second": self.reward_rate_per_second,
            "fuel_max_reserve": self.fuel_max_reserve,
            "food_max_reserve": self.food_max_reserve,
            "arms_max_reserve": self.arms_max_reserve,
            "toolkit_max_reserve": self.toolkit_max_reserve,
            "milliseconds_to_burn_one_fuel": self.milliseconds_to_burn_one_fuel,
            "milliseconds_to_burn_one_food": self.milliseconds_to_burn_one_food,
            "milliseconds_to_burn_one_arms": self.milliseconds_to_burn_one_arms,
            "milliseconds_to_burn_one_toolkit": self.milliseconds_to_burn_one_toolkit,
        }

    @classmethod
    def from_json(cls, obj: ScoreVarsShipJSON) -> "ScoreVarsShip":
        return cls(
            ship_mint=PublicKey(obj["ship_mint"]),
            reward_rate_per_second=obj["reward_rate_per_second"],
            fuel_max_reserve=obj["fuel_max_reserve"],
            food_max_reserve=obj["food_max_reserve"],
            arms_max_reserve=obj["arms_max_reserve"],
            toolkit_max_reserve=obj["toolkit_max_reserve"],
            milliseconds_to_burn_one_fuel=obj["milliseconds_to_burn_one_fuel"],
            milliseconds_to_burn_one_food=obj["milliseconds_to_burn_one_food"],
            milliseconds_to_burn_one_arms=obj["milliseconds_to_burn_one_arms"],
            milliseconds_to_burn_one_toolkit=obj["milliseconds_to_burn_one_toolkit"],
        )



import typing
from dataclasses import dataclass
from base64 import b64decode
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID
from ..types import order_side

class OrderAccountJSON(typing.TypedDict):
    order_initializer_pubkey: str
    currency_mint: str
    asset_mint: str
    initializer_currency_token_account: str
    initializer_asset_token_account: str
    order_side: order_side.OrderSideJSON
    price: int
    order_origination_qty: int
    order_remaining_qty: int
    created_at_timestamp: int


@dataclass
class OrderAccount:
    discriminator: typing.ClassVar = b"OCp\x9b\xd6\x0e 7"
    layout: typing.ClassVar = borsh.CStruct(
        "order_initializer_pubkey" / BorshPubkey,
        "currency_mint" / BorshPubkey,
        "asset_mint" / BorshPubkey,
        "initializer_currency_token_account" / BorshPubkey,
        "initializer_asset_token_account" / BorshPubkey,
        "order_side" / order_side.layout,
        "price" / borsh.U64,
        "order_origination_qty" / borsh.U64,
        "order_remaining_qty" / borsh.U64,
        "created_at_timestamp" / borsh.I64,
    )
    order_initializer_pubkey: PublicKey
    currency_mint: PublicKey
    asset_mint: PublicKey
    initializer_currency_token_account: PublicKey
    initializer_asset_token_account: PublicKey
    order_side: order_side.OrderSideKind
    price: int
    order_origination_qty: int
    order_remaining_qty: int
    created_at_timestamp: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.Optional["OrderAccount"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp["result"]["value"]
        if info is None:
            return None
        if info["owner"] != str(PROGRAM_ID):
            raise ValueError("Account does not belong to this program")
        bytes_data = b64decode(info["data"][0])
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.List[typing.Optional["OrderAccount"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["OrderAccount"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != PROGRAM_ID:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "OrderAccount":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = OrderAccount.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            order_initializer_pubkey=dec.order_initializer_pubkey,
            currency_mint=dec.currency_mint,
            asset_mint=dec.asset_mint,
            initializer_currency_token_account=dec.initializer_currency_token_account,
            initializer_asset_token_account=dec.initializer_asset_token_account,
            order_side=order_side.from_decoded(dec.order_side),
            price=dec.price,
            order_origination_qty=dec.order_origination_qty,
            order_remaining_qty=dec.order_remaining_qty,
            created_at_timestamp=dec.created_at_timestamp,
        )

    def to_json(self) -> OrderAccountJSON:
        return {
            "order_initializer_pubkey": str(self.order_initializer_pubkey),
            "currency_mint": str(self.currency_mint),
            "asset_mint": str(self.asset_mint),
            "initializer_currency_token_account": str(
                self.initializer_currency_token_account
            ),
            "initializer_asset_token_account": str(
                self.initializer_asset_token_account
            ),
            "order_side": self.order_side.to_json(),
            "price": self.price,
            "order_origination_qty": self.order_origination_qty,
            "order_remaining_qty": self.order_remaining_qty,
            "created_at_timestamp": self.created_at_timestamp,
        }

    @classmethod
    def from_json(cls, obj: OrderAccountJSON) -> "OrderAccount":
        return cls(
            order_initializer_pubkey=PublicKey(obj["order_initializer_pubkey"]),
            currency_mint=PublicKey(obj["currency_mint"]),
            asset_mint=PublicKey(obj["asset_mint"]),
            initializer_currency_token_account=PublicKey(
                obj["initializer_currency_token_account"]
            ),
            initializer_asset_token_account=PublicKey(
                obj["initializer_asset_token_account"]
            ),
            order_side=order_side.from_json(obj["order_side"]),
            price=obj["price"],
            order_origination_qty=obj["order_origination_qty"],
            order_remaining_qty=obj["order_remaining_qty"],
            created_at_timestamp=obj["created_at_timestamp"],
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class ProcessInitializeSellArgs(typing.TypedDict):
    price: int
    origination_qty: int


layout = borsh.CStruct("price" / borsh.U64, "origination_qty" / borsh.U64)


class ProcessInitializeSellAccounts(typing.TypedDict):
    order_initializer: PublicKey
    market_vars_account: PublicKey
    deposit_mint: PublicKey
    receive_mint: PublicKey
    order_vault_account: PublicKey
    order_vault_authority: PublicKey
    initializer_deposit_token_account: PublicKey
    initializer_receive_token_account: PublicKey
    order_account: PublicKey
    registered_currency: PublicKey
    open_orders_counter: PublicKey
    system_program: PublicKey
    rent: PublicKey
    token_program: PublicKey


def process_initialize_sell(
    args: ProcessInitializeSellArgs, accounts: ProcessInitializeSellAccounts
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["order_initializer"], is_signer=True, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["market_vars_account"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["deposit_mint"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["receive_mint"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["order_vault_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["order_vault_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["initializer_deposit_token_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["initializer_receive_token_account"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["order_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["registered_currency"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["open_orders_counter"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["rent"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["token_program"], is_signer=False, is_writable=False
        ),
    ]
    identifier = b"+*\xa7\xfc\x19/\xd4\xe1"
    encoded_args = layout.build(
        {
            "price": args["price"],
            "origination_qty": args["origination_qty"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class ProcessWithdrawFuelArgs(typing.TypedDict):
    staking_bump: int
    scorevars_bump: int
    scorevars_ship_bump: int
    escrow_auth_bump: int
    escrow_bump: int


layout = borsh.CStruct(
    "staking_bump" / borsh.U8,
    "scorevars_bump" / borsh.U8,
    "scorevars_ship_bump" / borsh.U8,
    "escrow_auth_bump" / borsh.U8,
    "escrow_bump" / borsh.U8,
)


class ProcessWithdrawFuelAccounts(typing.TypedDict):
    player_account: PublicKey
    ship_staking_account: PublicKey
    score_vars_account: PublicKey
    score_vars_ship_account: PublicKey
    fuel_token_account_escrow: PublicKey
    fuel_token_account_return: PublicKey
    fuel_mint: PublicKey
    escrow_authority: PublicKey
    token_program: PublicKey
    clock: PublicKey
    ship_mint: PublicKey


def process_withdraw_fuel(
    args: ProcessWithdrawFuelArgs, accounts: ProcessWithdrawFuelAccounts
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["player_account"], is_signer=True, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["ship_staking_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["score_vars_account"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["score_vars_ship_account"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["fuel_token_account_escrow"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["fuel_token_account_return"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["fuel_mint"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["escrow_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["token_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["clock"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["ship_mint"], is_signer=False, is_writable=False),
    ]
    identifier = b"-\x93\x93\xa3 \xdd\xa1\xb0"
    encoded_args = layout.build(
        {
            "staking_bump": args["staking_bump"],
            "scorevars_bump": args["scorevars_bump"],
            "scorevars_ship_bump": args["scorevars_ship_bump"],
            "escrow_auth_bump": args["escrow_auth_bump"],
            "escrow_bump": args["escrow_bump"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



from .market_vars import MarketVars, MarketVarsJSON
from .open_orders_counter import OpenOrdersCounter, OpenOrdersCounterJSON
from .order_account import OrderAccount, OrderAccountJSON
from .registered_currency import RegisteredCurrency, RegisteredCurrencyJSON



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class ProcessEnlistPlayerArgs(typing.TypedDict):
    bump: int
    faction_id: int


layout = borsh.CStruct("bump" / borsh.U8, "faction_id" / borsh.U8)


class ProcessEnlistPlayerAccounts(typing.TypedDict):
    player_faction_account: PublicKey
    player_account: PublicKey
    system_program: PublicKey
    clock: PublicKey


def process_enlist_player(
    args: ProcessEnlistPlayerArgs, accounts: ProcessEnlistPlayerAccounts
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["player_faction_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["player_account"], is_signer=True, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["clock"], is_signer=False, is_writable=False),
    ]
    identifier = b"\xc6j>\xa7w.\xcfQ"
    encoded_args = layout.build(
        {
            "bump": args["bump"],
            "faction_id": args["faction_id"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)



from setuptools import setup, find_packages
import os, pathlib
from setuptools.command.egg_info import egg_info

class egg_info_ex(egg_info):
    """Includes license file into `.egg-info` folder."""

    def run(self):
        # don't duplicate license into `.egg-info` when building a distribution
        if not self.distribution.have_run.get('install', True):
            # `install` command is in progress, copy license
            self.mkpath(self.egg_info)
            self.copy_file('LICENSE.txt', self.egg_info)

        egg_info.run(self)

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="staratlaspy",
    version="0.1.26",
    description="Python module for parsing staratlas instructions (based on anchorpy).",
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    readme = "README.md",
    license_files = ('LICENSE.txt',),
    license = 'Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License',
    cmdclass = {'egg_info': egg_info_ex},

    #url = "https://git.kryptonhub.com/the-club/staratlaspy",
    author = "Vlad Smirnov",
    author_email = "vlad@theclubguild.com",
    classifiers=[
      "Development Status :: 3 - Alpha",
      "Intended Audience :: Developers",
      "Topic :: Software Development :: Libraries :: Python Modules",
      "Programming Language :: Python :: 3"
    ],
    packages = ['staratlas'] +
               ['staratlas.utils'] +
               ['staratlas.marketplace'] +
               ['staratlas.marketplace.'+ s for s in ['accounts','errors','instructions','types']] +
               ['staratlas.faction'] +
               ['staratlas.faction.' + s for s in ['accounts', 'errors', 'instructions']] +
               ['staratlas.score'] +
               ['staratlas.score.' + s for s in ['accounts', 'errors', 'instructions', 'types']],
    package_dir = {'staratlas': '.'},
    python_requires=">=3.9, <4",
    install_requires=[
        'construct==2.10.67',
        'solders==0.2.0',
        'requests[socks]',
        'solana==0.23.1',
        'click==8.1.3',
        'anchorpy[cli]==0.9.1',
        'asyncclick==8.1.3.2'
    ],
    data_files=[("idls", [
        "idls/staratlas-score.json",
        "idls/staratlas-faction.json",
        "idls/staratlas-distributor.json",
        "idls/staratlas-marketplace.json"
    ])],
    include_package_data = True,
    entry_points={  # Optional
        "console_scripts": [
            "staratlas=staratlas.cli:cli",
        ],
    },
)




from .process_initialize import (
    process_initialize,
    ProcessInitializeArgs,
    ProcessInitializeAccounts,
)
from .process_register_ship import (
    process_register_ship,
    ProcessRegisterShipArgs,
    ProcessRegisterShipAccounts,
)
from .process_deregister_ship import (
    process_deregister_ship,
    ProcessDeregisterShipArgs,
    ProcessDeregisterShipAccounts,
)
from .process_update_reward_rate import (
    process_update_reward_rate,
    ProcessUpdateRewardRateArgs,
    ProcessUpdateRewardRateAccounts,
)
from .process_initial_deposit import (
    process_initial_deposit,
    ProcessInitialDepositArgs,
    ProcessInitialDepositAccounts,
)
from .process_partial_deposit import (
    process_partial_deposit,
    ProcessPartialDepositArgs,
    ProcessPartialDepositAccounts,
)
from .process_refuel import process_refuel, ProcessRefuelArgs, ProcessRefuelAccounts
from .process_refeed import process_refeed, ProcessRefeedArgs, ProcessRefeedAccounts
from .process_rearm import process_rearm, ProcessRearmArgs, ProcessRearmAccounts
from .process_repair import process_repair, ProcessRepairArgs, ProcessRepairAccounts
from .process_settle import process_settle, ProcessSettleArgs, ProcessSettleAccounts
from .process_harvest import process_harvest, ProcessHarvestArgs, ProcessHarvestAccounts
from .process_withdraw_fuel import (
    process_withdraw_fuel,
    ProcessWithdrawFuelArgs,
    ProcessWithdrawFuelAccounts,
)
from .process_withdraw_food import (
    process_withdraw_food,
    ProcessWithdrawFoodArgs,
    ProcessWithdrawFoodAccounts,
)
from .process_withdraw_arms import (
    process_withdraw_arms,
    ProcessWithdrawArmsArgs,
    ProcessWithdrawArmsAccounts,
)
from .process_withdraw_ships import (
    process_withdraw_ships,
    ProcessWithdrawShipsArgs,
    ProcessWithdrawShipsAccounts,
)
from .process_close_accounts import (
    process_close_accounts,
    ProcessCloseAccountsArgs,
    ProcessCloseAccountsAccounts,
)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class BuyJSON(typing.TypedDict):
    kind: typing.Literal["Buy"]


class SellJSON(typing.TypedDict):
    kind: typing.Literal["Sell"]


@dataclass
class Buy:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Buy"

    @classmethod
    def to_json(cls) -> BuyJSON:
        return BuyJSON(
            kind="Buy",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Buy": {},
        }


@dataclass
class Sell:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Sell"

    @classmethod
    def to_json(cls) -> SellJSON:
        return SellJSON(
            kind="Sell",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Sell": {},
        }


OrderSideKind = typing.Union[Buy, Sell]
OrderSideJSON = typing.Union[BuyJSON, SellJSON]


def from_decoded(obj: dict) -> OrderSideKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Buy" in obj:
        return Buy()
    if "Sell" in obj:
        return Sell()
    raise ValueError("Invalid enum object")


def from_json(obj: OrderSideJSON) -> OrderSideKind:
    if obj["kind"] == "Buy":
        return Buy()
    if obj["kind"] == "Sell":
        return Sell()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Buy" / borsh.CStruct(), "Sell" / borsh.CStruct())



import typing
from dataclasses import dataclass
from base64 import b64decode
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class MarketVarsJSON(typing.TypedDict):
    update_authority_master: str
    bump: int


@dataclass
class MarketVars:
    discriminator: typing.ClassVar = b"\xff\x8e\x86\x198\x01\xdb|"
    layout: typing.ClassVar = borsh.CStruct(
        "update_authority_master" / BorshPubkey, "bump" / borsh.U8
    )
    update_authority_master: PublicKey
    bump: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.Optional["MarketVars"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp["result"]["value"]
        if info is None:
            return None
        if info["owner"] != str(PROGRAM_ID):
            raise ValueError("Account does not belong to this program")
        bytes_data = b64decode(info["data"][0])
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.List[typing.Optional["MarketVars"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["MarketVars"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != PROGRAM_ID:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "MarketVars":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = MarketVars.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            update_authority_master=dec.update_authority_master,
            bump=dec.bump,
        )

    def to_json(self) -> MarketVarsJSON:
        return {
            "update_authority_master": str(self.update_authority_master),
            "bump": self.bump,
        }

    @classmethod
    def from_json(cls, obj: MarketVarsJSON) -> "MarketVars":
        return cls(
            update_authority_master=PublicKey(obj["update_authority_master"]),
            bump=obj["bump"],
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class InitializeOpenOrdersCounterAccounts(typing.TypedDict):
    user: PublicKey
    open_orders_counter: PublicKey
    deposit_mint: PublicKey
    system_program: PublicKey


def initialize_open_orders_counter(
    accounts: InitializeOpenOrdersCounterAccounts,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["user"], is_signer=True, is_writable=True),
        AccountMeta(
            pubkey=accounts["open_orders_counter"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["deposit_mint"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["system_program"], is_signer=False, is_writable=False
        ),
    ]
    identifier = b"\xdd\x86\x05L\x04\x91\xca\x1d"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, PROGRAM_ID, data)
